import numpy as np

from ifra.predictor import ClassificationPredictor
from ruleskit import RuleSet, ClassificationRule, HyperrectangleCondition
import pandas as pd

x_data = pd.DataFrame(
    index=[2, 4, 6, 8, 10, 12, 14, 16, 18, 20],
    columns=["X1", "X2", "X3", "X4"],
    data=[
        [0, 10, 100, 1000],
        [1, 10, 100, 1000],
        [0, 10, 100, 1000],
        [5, 10, 100, 1000],
        [6, 10, 100, 1000],
        [5, 10, 100, 1000],
        [0, 10, 160, 1000],
        [0, 10, 150, 1000],
        [6, 10, 150, 1000],
        [5, 10, 160, 1000],
    ]
)

condition_0 = HyperrectangleCondition(features_names=["X1"], features_indexes=[0], bmins=[0], bmaxs=[4])
predictions_0_int = np.array([2, 2, 2, np.nan, np.nan, np.nan, 2, 2, np.nan, np.nan])
condition_100 = HyperrectangleCondition(features_names=["X3"], features_indexes=[2], bmins=[150], bmaxs=[170])
predictions_100_int = np.array([np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, 3, 3, 3, 3])
condition_10_100 = HyperrectangleCondition(features_names=["X1", "X3"], features_indexes=[0, 2], bmins=[0, 150], bmaxs=[4, 170])
predictions_10_100_int = np.array([np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, 5, 5, np.nan, np.nan])



from ifra.central_server import CentralServer, NodeGate
from ifra.configs import Config, CentralConfig, NodePublicConfig
from ifra.messenger import NodeMessenger


def one_node(node):
    assert isinstance(node, NodeGate)
    assert isinstance(node.public_configs, (Config, NodePublicConfig))
    assert isinstance(node.messenger, NodeMessenger)
    assert node.messenger.running is False
    assert node.messenger.fitting is False
    assert node.messenger.stop is False
    assert node.messenger.error is None
    assert node.messenger.central_error is None
    assert node.ruleset is None
    assert node.last_fetch is None
    assert node.new_data is False
    assert node.checked is False
    assert node.checked_once is False
    assert node.ok is True


def test_simple_init_and_watch():
    server = CentralServer(
        nodes_configs_paths=[
            "tests/data/node_0/public_configs.json",
            "tests/data/node_1/public_configs.json",
            "tests/data/node_2/public_configs.json",
            "tests/data/node_3/public_configs.json",
        ],
        central_configs_path="tests/data/central_configs.json",
    )
    assert isinstance(server.central_configs, (Config, CentralConfig))
    assert len(server.nodes) == 4
    for node in server.nodes:
        one_node(node)



from ifra import Node
from ifra.configs import Config, NodePublicConfig, NodeDataConfig
from ifra.messenger import NodeMessenger
from ifra.fitters import Fitter, DecisionTreeFitter
from ifra.updaters import Updater, AdaBoostUpdater
from ifra.datapreps import DataPrep, BinFeaturesDataPrep
from ..alternates.dataprep import AlternateDataPrep
from ..alternates.updater import AlternateUpdater
from ..alternates.fitter import AlternateFitter


def test_init_and_watch_simple(clean):
    node = Node("tests/data/node_test/public_configs.json", "tests/data/node_test/data_configs.json")
    assert isinstance(node.public_configs, (Config, NodePublicConfig))
    assert isinstance(node.data, (Config, NodeDataConfig))
    assert isinstance(node.messenger, NodeMessenger)
    assert node.messenger.running is False
    assert node.messenger.fitting is False
    assert node.messenger.stop is False
    assert node.messenger.error is None
    assert node.messenger.central_error is None
    assert isinstance(node.fitter, (Fitter, DecisionTreeFitter))
    assert isinstance(node.updater, (Updater, AdaBoostUpdater))
    assert isinstance(node.dataprep, (DataPrep, BinFeaturesDataPrep))
    assert node.datapreped is False
    assert node.data.dataprep_kwargs == node.public_configs.dataprep_kwargs
    assert node.copied is False
    assert node.ruleset is None
    assert node.last_fetch is None

    assert not (node.data.x_path.parent / "x_datapreped.csv").is_file()
    assert not (node.data.y_path.parent / "y_datapreped.csv").is_file()
    assert not (node.data.x_path.parent / "x_datapreped_copy_for_learning.csv").is_file()
    assert not (node.data.y_path.parent / "y_datapreped_copy_for_learning.csv").is_file()
    assert not (node.data.x_path.parent / "plots").is_dir()
    assert not (node.data.x_path.parent / "plots_datapreped").is_dir()

    node.watch(timeout=5, sleeptime=1)

    assert node.copied
    assert node.datapreped
    assert node.messenger.fitting is False
    assert node.messenger.stop is True
    assert (node.data.x_path.parent / "x_datapreped.csv").is_file()
    assert (node.data.y_path.parent / "y_datapreped.csv").is_file()
    assert (node.data.x_path.parent / "x_datapreped_copy_for_learning.csv").is_file()
    assert (node.data.y_path.parent / "y_datapreped_copy_for_learning.csv").is_file()
    assert (node.data.x_path.parent / "plots").is_dir()
    assert (node.data.x_path.parent / "plots_datapreped").is_dir()
    assert node.public_configs.local_model_path.is_file()


def test_init_alternate_dataprep_updater_fitter():
    node = Node("tests/data/node_test/alternate_public_configs.json", "tests/data/node_test/data_configs.json")
    assert isinstance(node.dataprep, AlternateDataPrep)
    assert isinstance(node.updater, AlternateUpdater)
    assert isinstance(node.fitter, AlternateFitter)



import pandas as pd
from ruleskit import RuleSet
from typing import Tuple
from ifra.updaters import Updater


class AlternateUpdater(Updater):
    @staticmethod
    def make_update(x: pd.DataFrame, y: pd.DataFrame, ruleset: RuleSet) -> Tuple[pd.DataFrame, pd.DataFrame]:
        ruleset.remember_activation = True
        ruleset.calc_activation(x.values)
        x = x[ruleset.activation == 1]
        y = y[ruleset.activation == 1]
        return x, y



from ifra.aggregations import aggregate_classif_preds
from ruleskit import RuleSet, ClassificationRule, HyperrectangleCondition
import pytest


predictions_float = [1, 1, 1.1, 2, 2, 2.1, 2.1, 3]
predictions_str = ["a", "a", "aa", "b", "b", "bb", "bb", "c"]

ruleset_float = RuleSet(
    [
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 3])
        ),
    ]
)

for i in range(len(ruleset_float)):
    # noinspection PyProtectedMember
    ruleset_float[i]._prediction = predictions_float[i]

ruleset_str = RuleSet(
    [
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 3])
        ),
    ]
)

for i in range(len(ruleset_str)):
    # noinspection PyProtectedMember
    ruleset_str[i]._prediction = predictions_str[i]

expected_ruleset_float = RuleSet(
    [
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 3])
        ),
    ]
)

expected_ruleset_str = RuleSet(
    [
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 3])
        ),
    ]
)

# noinspection PyProtectedMember
expected_ruleset_float[0]._prediction = 1
# noinspection PyProtectedMember
expected_ruleset_float[1]._prediction = 1
# noinspection PyProtectedMember
expected_ruleset_float[2]._prediction = 1
# noinspection PyProtectedMember
expected_ruleset_float[3]._prediction = 3
# noinspection PyProtectedMember

expected_ruleset_str[0]._prediction = "a"
# noinspection PyProtectedMember
expected_ruleset_str[1]._prediction = "a"
# noinspection PyProtectedMember
expected_ruleset_str[2]._prediction = "a"
# noinspection PyProtectedMember
expected_ruleset_str[3]._prediction = "c"
# noinspection PyProtectedMember


@pytest.mark.parametrize(
    "ruleset, expected_ruleset", ((ruleset_str, expected_ruleset_str), (ruleset_float, expected_ruleset_float))
)
def test_classif_pred_aggr(ruleset, expected_ruleset):

    aggregate_classif_preds(ruleset)

    assert len(ruleset) == len(expected_ruleset)
    for rule, exprule in zip(ruleset, expected_ruleset):
        assert rule == exprule
        assert rule.condition == exprule.condition
        assert rule.prediction == exprule.prediction



from typing import List
from ruleskit import RuleSet, RegressionRule, ClassificationRule
import logging
from collections import Counter

logger = logging.getLogger(__name__)

# TODO (pcotte) : make it a watcher for data separation


class Aggregation:
    """Abstract class for aggregating node models."""

    # noinspection PyUnresolvedReferences
    def __init__(self, central_server: "CentralServer", **kwargs):
        """
        Parameters
        ----------
        central_server: CentralServer
            `ifra.central_server.CentralServer` object
        kwargs:
            Any additionnal keyword argument that the overleading class accepts. Those arguments will become attributes.
        """
        self.central_server = central_server
        for arg in kwargs:
            setattr(self, arg, kwargs[arg])

    def aggregate(self, rulesets: List[RuleSet]) -> str:
        """Aggregates the node models into central model, then updates the central model's rules predictions

        Parameters
        ----------
        rulesets: List[RuleSet]
            New rulesets provided by the nodes.

        Returns
        -------
        str\n
          * "updated" if central model was updates\n
          * "pass" if no new rules were found but learning should continue\n
          * "stop" otherwise. Will stop learning.\n"""
        result = self._aggregate(rulesets)
        self.aggregate_predictions()
        return result

    def _aggregate(self, rulesets: List[RuleSet]) -> str:
        """To be implemented in daughter class.

         Must modify inplace `ifra.central_server.CentralServer` *ruleset*.

        Parameters
        ----------
        rulesets: List[RuleSet]
            New rulesets provided by the nodes.

        Returns
        -------
        str\n
          * "updated" if central model was updates\n
          * "pass" if no new rules were found but learning should continue\n
          * "stop" otherwise. Will stop learning.\n
        """
        pass

    def aggregate_predictions(self) -> None:
        """To be implemented in daughter class.

         Must modify inplace the rules predictions in `ifra.central_server.CentralServer` *ruleset*
        """
        pass


class AdaBoostAggregation(Aggregation):
    """Among all rules generated by the nodes in the current iteration, will keep only the most recurent rule(s).

    Can be used by giving *adaboost_aggregation* as *aggregation* argument when creating a
    `ifra.central_server.CentralServer`

    Those rules will be added to  the given ruleset, returned by the function,
    and the points in each node's data activated by it will be ignored in order to find other relevant rules.
    In order not to remove too many points, a maximum of coverage is imposed.

    If among the rules extracted from the nodes, none are new compared to the current state of the model, nodes are
    not updated. If no new rules are found but each nodes had new data, learning is finished.
    """

    def _aggregate(self, rulesets: List[RuleSet]) -> str:
        logger.info("Aggregating fit results using AdaBoost method...")
        all_rules = []
        n_nodes = len(self.central_server.nodes)

        for rs in rulesets:
            all_rules += rs.rules

        if len(all_rules) == 0:
            logger.info("... no rules found")
            if len(rulesets) == n_nodes:
                logger.info("No new rules were found, despite all nodes having provided a new model : learning is over")
                return "stop"
            return "pass"

        if self.central_server.ruleset is None:
            occurences = {r: all_rules.count(r)for r in set(all_rules)}
        else:
            occurences = {
                r: all_rules.count(r)
                for r in set(all_rules)
                if r not in self.central_server.ruleset
            }

        if len(occurences) == 0:
            logger.warning("No new rules matched coverage criterion")
            if len(rulesets) == n_nodes:
                logger.info("No new rules were found, despite all nodes having provided a new model : learning is over")
                return "stop"
            return "pass"
        max_occurences = max(list(occurences.values()))

        if self.central_server.ruleset is None:
            new_rules = RuleSet(remember_activation=False)
        else:
            new_rules = RuleSet(rules_list=self.central_server.ruleset.rules, remember_activation=False)

        found_new = 0
        for r in occurences:
            if r not in new_rules and occurences[r] == max_occurences:
                found_new += 1
                new_rules.append(r, update_activation=False)

        if found_new == 0:
            logger.warning("No new rules found")
            if len(rulesets) == n_nodes:
                logger.info(
                    "No new rules were found, despite all nodes having provided a new model" " : learning is over"
                )
                return "stop"
            return "pass"
        else:
            logger.info(f"Aggregated {found_new} new rules")

        # new_rules.check_duplicated_rules(
        #     new_rules.rules, name_or_index="name" if len(new_rules.features_names) > 0 else "index"
        # )

        del self.central_server.ruleset
        self.central_server.ruleset = new_rules
        return "updated"

    def aggregate_predictions(self):
        if self.central_server.ruleset.rule_type == ClassificationRule:
            aggregate_classif_preds(self.central_server.ruleset)
        elif self.central_server.ruleset.rule_type == RegressionRule:
            aggregate_regr_preds(self.central_server.ruleset)
        else:
            raise TypeError(f"Unexpected ruleset's rule type {self.central_server.ruleset.rule_type}")


def aggregate_classif_preds(ruleset: RuleSet):
    """For now, the ruleset can contain rules with identical conditions but different predictions.
    This method aims at solving that : for duplicated conditions, the corresponding rules will have their
    predictions set to be the one of the most recurring prediction. If several predictions are equally frequent, the
    rules corresponding to the duplicated conditions are discarder.

    Parameters
    ----------
    ruleset: RuleSet
        Modified in place
    """
    duplicated_conditions = {}
    to_remove = []
    conditions = [r.condition for r in ruleset]

    if list(set(conditions)) == conditions:  # No duplicated conds -> no contradictory predictions -> do nothing
        return

    for i in range(len(ruleset)):
        if conditions.count(conditions[i]) > 1:
            if conditions[i] not in duplicated_conditions:
                duplicated_conditions[conditions[i]] = [i]
            else:
                duplicated_conditions[conditions[i]].append(i)

    for condition in duplicated_conditions:
        preds = []
        for i in duplicated_conditions[condition]:
            preds.append(ruleset[i].prediction)
        if len(set(preds)) == 1:  # Duplicated conditions predict the same : nothing to do
            continue
        counter = Counter(preds).most_common(2)
        """Is a list of 2 tuples containing the 2 most frequent elements in *preds* with the number of times they
         are present"""
        if counter[0][1] == counter[1][1]:  # 2 or more different predictions are equally frequent : ignore rules
            to_remove += duplicated_conditions[condition]
            continue
        good_pred = counter[0][0]
        for i in duplicated_conditions[condition]:
            # noinspection PyProtectedMember
            ruleset[i]._prediction = good_pred

    ruleset._rules = [ruleset[i] for i in range(len(ruleset)) if i not in to_remove]
    ruleset.remember_activation = False
    ruleset.stack_activation = False
    ruleset._activation = None
    ruleset.stacked_activations = None
    ruleset._coverage = None


def aggregate_regr_preds(self):
    # TODO (pcotte)
    pass
